using System.Collections;
using System.Collections.Generic;
using UnityEngine;

// Matrix is used in DirectX and OpenGL, and also used in Unity shader code (revised 2024/10/10 by Dankook University Kyoung Shin Park)
public class MatrixTest : MonoBehaviour
{
    public static void Print2DArray(Matrix4x4 matrix)
    {
        for (int i = 0; i < 4; i++)
        {
            Debug.Log(matrix[i,0] + "\t" + matrix[i,1] + "\t" + matrix[i,2] + "\t" + matrix[i,3]); // row, column
        }
    }

    public static Matrix4x4 View(Vector3 pos, Vector3 look, Vector3 up, Vector3 right) 
    {
        // keep camera's axes orthogonal to each other
        look = Vector3.Normalize(look);
        up = Vector3.Normalize(up);

        right = Vector3.Normalize(Vector3.Cross(up, look));
        Debug.Log("right=" + right);

        up = Vector3.Normalize(Vector3.Cross(look, right));
        Debug.Log("up=" + up);

        Matrix4x4 V = Matrix4x4.identity;
        V.SetColumn(0, new Vector4(right.x, right.y, right.z, -Vector3.Dot(right, pos)));
        V.SetColumn(1, new Vector4(up.x, up.y, up.z, -Vector3.Dot(up, pos)));
        V.SetColumn(2, new Vector4(look.x, look.y, look.z, -Vector3.Dot(look, pos)));
        V.SetColumn(3, new Vector4(0.0f, 0.0f, 0.0f, 1.0f));

        Debug.Log("V=\n" + V);
        //for (int i = 0; i < 4; i++) Debug.Log(V.GetRow(i));

        return V;
    }

    // https://paulbourke.net/geometry/rotate/
    /*
    Rotate a point p by angle around an arbitrary line segment p1-p2
    Return the rotated point.
    Positive angles are anticlockwise looking down the axis
    towards the origin. Assume right hand coordinate system.  
    */
    public static Vector3 ArbitraryRotate(Vector3 p, float angle, Vector3 p1, Vector3 p2)
    {
        Vector3 u, q1, q2, q3, q4, q5, q6, q7;
        float d;

        // (1) translate space so that the rotation axis passes through the origin
        q1 = p - p1;
        u = p2 - p1; // rotation-axis vector
        u = Vector3.Normalize(u);
        d = Mathf.Sqrt(u.y*u.y + u.z*u.z);

        // (2) rotate space about the x axis so that the rotation axis lies in the xz plane
        if (d != 0) {
            q2.x = q1.x;
            q2.y = q1.y * u.z / d - q1.z * u.y / d;
            q2.z = q1.y * u.y / d + q1.z * u.z / d;
        } else {
            q2 = q1;
        }

        // (3) rotate space about the y axis so that the rotation axis lies along the z axis
        q3.x = q2.x * d - q2.z * u.x;
        q3.y = q2.y;
        q3.z = q2.x * u.x + q2.z * d;

        // (4) perform the desired rotation by theta about the z axis
        float theta = Mathf.Deg2Rad * angle;
        q4.x = q3.x * Mathf.Cos(theta) - q3.y * Mathf.Sin(theta);
        q4.y = q3.x * Mathf.Sin(theta) + q3.y * Mathf.Cos(theta);
        q4.z = q3.z;

        // (5) apply the inverse of step (3)
        q5.x =  q4.x * d + q4.z * u.x;
        q5.y =  q4.y;
        q5.z = -q4.x * u.x + q4.z * d;

        // (6) apply the inverse of step (2)
        if (d != 0) {
            q6.x =  q5.x;
            q6.y =  q5.y * u.z / d + q5.z * u.y / d;
            q6.z = -q5.y * u.y / d + q5.z * u.z / d;
        } else {
            q6 = q5;
        }

        // (7) apply the inverse of step (1)
        q7 = q6 + p1;
        return q7;
    }

    /*
    Rotate a point p by angle around an arbitrary axis
    Return the rotated point.
    Positive angles are anticlockwise looking down the axis
    towards the origin. Assume right hand coordinate system.
    */
    public static Vector3 ArbitraryRotate(Vector3 p, float angle, Vector3 axis)
    {
        Vector3 q = Vector3.zero;
        float costheta, sintheta;
        float theta = Mathf.Deg2Rad * angle;

        axis = Vector3.Normalize(axis);
        costheta = Mathf.Cos(theta);
        sintheta = Mathf.Sin(theta);

        q.x += (costheta + (1 - costheta) * axis.x * axis.x) * p.x;
        q.x += ((1 - costheta) * axis.x * axis.y - axis.z * sintheta) * p.y;
        q.x += ((1 - costheta) * axis.x * axis.z + axis.y * sintheta) * p.z;

        q.y += ((1 - costheta) * axis.x * axis.y + axis.z * sintheta) * p.x;
        q.y += (costheta + (1 - costheta) * axis.y * axis.y) * p.y;
        q.y += ((1 - costheta) * axis.y * axis.z - axis.x * sintheta) * p.z;

        q.z += ((1 - costheta) * axis.x * axis.z - axis.y * sintheta) * p.x;
        q.z += ((1 - costheta) * axis.y * axis.z + axis.x * sintheta) * p.y;
        q.z += (costheta + (1 - costheta) * axis.z * axis.z) * p.z;

        return q;
    }

    // rotation using arbitrary axis and angle
    public static Matrix4x4 RotateAngleAxis(float angle, Vector3 axis)
    {
        Matrix4x4 M = Matrix4x4.identity;

        float theta = Mathf.Deg2Rad * angle;
        axis = Vector3.Normalize(axis);
        float costheta = Mathf.Cos(theta);
        float sintheta = Mathf.Sin(theta);
        
        M[0]  = costheta + (1 - costheta) * axis.x * axis.x;
        M[1]  = (1 - costheta) * axis.x * axis.y + sintheta * axis.z;
        M[2]  = (1 - costheta) * axis.x * axis.z - sintheta * axis.y;
        M[3]  = 0;
        
        M[4]  = (1 - costheta) * axis.y * axis.x - sintheta * axis.z;
        M[5]  = costheta + (1 - costheta) * axis.y * axis.y;
        M[6]  = (1 - costheta) * axis.y * axis.z + sintheta * axis.x;
        M[7]  = 0;
        
        M[8]  = (1 - costheta) * axis.z * axis.x + sintheta * axis.y;
        M[9]  = (1 - costheta) * axis.z * axis.y - sintheta * axis.x;
        M[10] = costheta + (1 - costheta) * axis.z * axis.z;
        M[11] = 0;
        
        M[12] = 0;
        M[13] = 0;
        M[14] = 0;
        M[15] = 1;

        return M;
    }

    private void Awake()
    {
        // matrix test
        Matrix4x4 M = new Matrix4x4();
        M.SetColumn(0, new Vector4(1.0f, 2.0f, 3.0f, 4.0f)); // column1
        M.SetColumn(1, new Vector4(5.0f, 6.0f, 7.0f, 8.0f)); // column2
	    M.SetColumn(2, new Vector4(9.0f, 10.0f, 11.0f, 12.0f)); // column3
	    M.SetColumn(3, new Vector4(13.0f, 14.0f, 15.0f, 16.0f)); // column4
        Debug.Log("M =\n" + M);

        Matrix4x4 A = new Matrix4x4();
        A.SetColumn(0, new Vector4(1.0f, 0.0f, 0.0f, 0.0f)); // column1
        A.SetColumn(1, new Vector4(0.0f, 2.0f, 0.0f, 0.0f)); // column2
	    A.SetColumn(2, new Vector4(0.0f, 0.0f, 4.0f, 0.0f)); // column3
	    A.SetColumn(3, new Vector4(1.0f, 2.0f, 3.0f, 1.0f)); // column4
        Debug.Log("A =\n" + A);

        Matrix4x4 B = new Matrix4x4();
        B.SetColumn(0, new Vector4(1.0f, 0.0f, 0.0f, 0.0f)); // column1
        B.SetColumn(1, new Vector4(0.0f, 1.0f, 0.0f, 0.0f)); // column2
	    B.SetColumn(2, new Vector4(0.0f, 0.0f, 1.0f, 0.0f)); // column3
	    B.SetColumn(3, new Vector4(2.0f, 2.0f, 2.0f, 1.0f)); // column4
        Debug.Log("B =\n" + B);

        Matrix4x4 C = A * B; // multiplication
        Debug.Log("C = A * B =\n" + C);

        Matrix4x4 D = B * A; // multiplication
        Debug.Log("D = B * A =\n" + D);

        Matrix4x4 E = A.inverse;		// inverse of the matrix
        Debug.Log("E = inverse(A) =\n" + E);

        Matrix4x4 I = A * E;			// identity matrix
        Debug.Log("I = A * E =\n" + I);

        Vector3 pos = new Vector3(1.0f, 0.0f, 0.0f); 
        Debug.Log("pos = " + pos);
        Vector3 q = A.MultiplyPoint3x4(pos);
        Debug.Log("q = A*pos = " + q);
        Vector3 r = B.MultiplyPoint3x4(pos);
        Debug.Log("r = B*pos = " + r);
        Vector3 s = C.MultiplyPoint3x4(pos);
        Debug.Log("s = C*pos = " + s);
        Vector3 t = D.MultiplyPoint3x4(pos);
        Debug.Log("t = D*pos = " + t);

        // translation matrix
        Matrix4x4 Tx = Matrix4x4.Translate(new Vector3(2.0f, 0.0f, 0.0f));
        Debug.Log("Tx =\n" + Tx);
        Matrix4x4 Ty = Matrix4x4.Translate(new Vector3(0.0f, 2.0f, 0.0f));
        Debug.Log("Ty =\n" + Ty);
        Matrix4x4 Tz = Matrix4x4.Translate(new Vector3(0.0f, 0.0f, 2.0f));
        Debug.Log("Tz =\n" + Tz);
        Vector3 tV = Tx.MultiplyPoint3x4(pos);
        Debug.Log("tV = Tx*pos =" + tV);

        // rotation matrix
        Matrix4x4 Rz = Matrix4x4.Rotate(Quaternion.AngleAxis(45.0f, new Vector3(0, 0, 1)));
        Debug.Log("Rz =\n" + Rz);
        Matrix4x4 Ra = RotateAngleAxis(45.0f, new Vector3(1, 1, 1));
        Debug.Log("Ra =\n" + Ra);
        Matrix4x4 Ra2 = Matrix4x4.Rotate(Quaternion.AngleAxis(45f, new Vector3(1, 1, 1)));
        Debug.Log("Quaternion.AngleAxis Ra2=\n" + Ra2);
        if (Ra == Ra2) Debug.Log("Ra == Ra2");
        Matrix4x4 Ra3 = Matrix4x4.Rotate(Quaternion.Euler(45f, 45f, 45f));
        Debug.Log("Quaternion.Euler Ra3=\n" + Ra3);
        if (Ra != Ra3) Debug.Log("Ra != Ra3");
        Vector3 tV1 = Ra.MultiplyPoint3x4(pos);
        Debug.Log("tV1 = Ra*pos = " + tV1);

        // scaling matrix
        Matrix4x4 Sy = Matrix4x4.Scale(new Vector3(1.0f, 2.0f, 1.0f));
        Debug.Log("Sy =\n" + Sy);
        Vector3 tV2 = Sy.MultiplyPoint3x4(pos);
        Debug.Log("tV2 = Sy*pos = " + tV2);

        Matrix4x4 TR = Tx * Rz; // Rotate Z Axis and then Translate X
        Debug.Log("TR = Tx * Rz =\n" + TR);
        Vector3 tV3 = TR.MultiplyPoint3x4(pos);
        Debug.Log("tV3 = TR*pos = " + tV3);

        Matrix4x4 RT = Rz * Tx; // Translate X and then Rotate Z Axis
        Debug.Log("RT = Rz * Tx =\n" + RT);
        Vector3 tV4 = RT.MultiplyPoint3x4(pos);
        Debug.Log("tV4 = RT*pos = " + tV4);

        Vector3 p = new Vector3(2.0f, 2.0f, 2.0f);
        Debug.Log("p=" + p);
        Vector3 p1 = ArbitraryRotate(p, 45.0f, new Vector3(1f, 2f, 3f), new Vector3(2f, 4f, 6f));
        Debug.Log("ArbitraryRotate p1=" + p1);
        Vector3 p2 = ArbitraryRotate(p, 45.0f, new Vector3(1f, 2f, 3f));
        Debug.Log("ArbitraryRotate p2=" + p2);
        if (p1 == p2) Debug.Log("p1 == p2");
        Vector3 p3 = Matrix4x4.Rotate(Quaternion.AngleAxis(45f, new Vector3(1f, 2f, 3f))).MultiplyPoint3x4(p);
        Debug.Log("Quaternion.AngleAxis p3=" + p3);
        if (p1 == p3) Debug.Log("p1 == p3");
        Vector3 p4 = RotateAngleAxis(45f, new Vector3(1f, 2f, 3f)).MultiplyPoint3x4(p);
        Debug.Log("RotateAngleAxis p4=" + p4);
        if (p1 == p4) Debug.Log("p1 == p4");

        // typical unity transformation
        Vector3 scale = new Vector3(0.2f, 0.2f, 0.2f);
        Quaternion rotation = Quaternion.Euler(0, 45f, 0);
        Vector3 translation = new Vector3(1f, 2f, -3f);

        Matrix4x4 s1 = Matrix4x4.Scale(scale);
        Debug.Log("s1=\n" + s1);
        //for (int i = 0; i < 4; i++) Debug.Log(s.GetRow(i));
        Matrix4x4 r1 = Matrix4x4.Rotate(rotation);
        Debug.Log("r1=\n" + r1);
        //for (int i = 0; i < 4; i++) Debug.Log(r.GetRow(i));
        Matrix4x4 t1 = Matrix4x4.Translate(translation);
        Debug.Log("t1=\n" + t1);
        //for (int i = 0; i < 4; i++) Debug.Log(t.GetRow(i));

        // create a composing transformation Scale -> Rotation -> Translation
        Matrix4x4 m1 = Matrix4x4.TRS(translation, rotation, scale);
        Debug.Log("m1=\n" + m1);
        //for (int i = 0; i < 4; i++) Debug.Log(m1.GetRow(i));

        // create a composing transformation Scale -> Rotation -> Translation
        Matrix4x4 m2 = Matrix4x4.Translate(translation) * Matrix4x4.Rotate(rotation) * Matrix4x4.Scale(scale);
        Debug.Log("m2=\n" + m2);
        //for (int i = 0; i < 4; i++) Debug.Log(m2.GetRow(i));

        if (m1 == m2) Debug.Log("m1 == m2");

        // create a composing transformation Translation -> Rotation -> Scale
        Matrix4x4 m3 = Matrix4x4.Scale(scale) * Matrix4x4.Rotate(rotation) * Matrix4x4.Translate(translation);
        Debug.Log("m3=\n" + m3);
        //for (int i = 0; i < 4; i++) Debug.Log(m3.GetRow(i));

        if (m2 != m3) Debug.Log("m2 != m3");

        // get a new position by a composing transformation matrix, m
        Vector3 position = new Vector3(5, 0, 0);
        Vector3 newposition = m1.MultiplyPoint(position);
        Debug.Log("newposition=" + newposition);
        Vector3 newposition2 = m2.MultiplyPoint(position);
        Debug.Log("newposition2=" + newposition2);
        Vector3 newposition3 = m3.MultiplyPoint(position);
        Debug.Log("newposition3=" + newposition3);

        // create a composing transformation Euler(30, 60, 45)
        Matrix4x4 rx = Matrix4x4.Rotate(Quaternion.Euler(30.0f, 0, 0));
        Debug.Log("rx=\n" + rx);
        //for (int i = 0; i < 4; i++) Debug.Log(rx.GetRow(i));
        Matrix4x4 ry = Matrix4x4.Rotate(Quaternion.Euler(0, 60.0f, 0));
        Debug.Log("ry=\n" + ry);
        //for (int i = 0; i < 4; i++) Debug.Log(ry.GetRow(i));
        Matrix4x4 rz = Matrix4x4.Rotate(Quaternion.Euler(0, 0, 45.0f));
        Debug.Log("rz=\n" + rz);
        //for (int i = 0; i < 4; i++) Debug.Log(rz.GetRow(i));
        Matrix4x4 ra = Matrix4x4.Rotate(Quaternion.Euler(30.0f, 60.0f, 45.0f));
        Debug.Log("ra=\n" + ra);
        //for (int i = 0; i < 4; i++) Debug.Log(ra.GetRow(i));
        // create a composing transformation rz -> rx -> ry
        // If using Euler angles, they're applied as rotations about the world axes in the order:
        // Roll (rotation about the world z axis)
        // Pitch (rotation about the world x axis)
        // Yaw (rotation about the world y axis)
        Matrix4x4 rb = ry * rx * rz;
        Debug.Log("rb=\n" + rb);
        //for (int i = 0; i < 4; i++) Debug.Log(rb.GetRow(i));
        if (ra == rb) Debug.Log("ra == rb");
        // create a composing transformation ry -> rx -> rz
        Matrix4x4 rc = rz * rx * ry;
        Debug.Log("rc=\n" + rc);
        //for (int i = 0; i < 4; i++) Debug.Log(rc.GetRow(i));
        if (rb != rc) Debug.Log("rb != rc");
    }

    // Start is called before the first frame update
    void Start()
    {
        
    }
/*
    // Update is called once per frame
    void Update()
    {
        // Extract TRS from transform.localToWorldMatrix
        Matrix4x4 m = transform.localToWorldMatrix;
 
        // Get translation, rotation, scale from matrix
        Vector3 t2 = m.GetColumn(3);
        Debug.Log("Translation=" + t2);
 
        Quaternion r2 = Quaternion.LookRotation(
            m.GetColumn(2),
            m.GetColumn(1)
        ); 
        Debug.Log("Rotation=" + r2.eulerAngles);
 
        Vector3 s2 = new Vector3(
            m.GetColumn(0).magnitude,
            m.GetColumn(1).magnitude,
            m.GetColumn(2).magnitude
        );
        Debug.Log("Scale=" + s2);
    }
*/
}
